/*
 * row.h
 *
 * Copyright (C) 2009 Tim Marston <edam@waxworlds.org>
 *
 * This file is part of sqlite3cc (hereafter referred to as "this program").
 * See http://www.waxworlds.org/edam/software/sqlite3cc for more information.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef SQLITE3CC_ROW_H_
#define SQLITE3CC_ROW_H_


#include <sqlite3cc/query.h>
#include <sqlite3.h>
#include <boost/utility.hpp>
#include <boost/lexical_cast.hpp>
#include <boost/utility/value_init.hpp>
#include <cassert>
#include <iostream>


namespace sqlite
{


struct _null_t;
struct _set_index_t;


/**
 * A result row from a query.
 *
 * The row is only valid until the next call to step() or reset() on the parent
 * query object, or until the parent query object is destructed. This may change
 * in future versions.
 */
class row
{
//______________________________________________________________________________
//                                                                 instantiation
protected:

	/**
	 * Constructor that provides the database that we were extracted from.
	 * @param the query that this row belongs to
	 * @oaram true if this row represents there being no more rows
	 */
	explicit row(
		query &query,
		bool valid = true );

//______________________________________________________________________________
//                                                              public interface
public:

	/**
	 * Determine if this row is valid or not. If it is not valid, there are no
	 * more rows in the results of the query.
	 */
	inline operator bool()
	{
		return _valid;
	}

	/**
	 * Ascertain a column's type.
	 * @param index column index
	 * @return sqlite datatype code
	 * @see sqlite3_column_type()
	 */
	int column_type(
		unsigned int index );

	/**
	 * get the number of bytes in the result for a given column.
	 * @param index column index
	 * @return number of bytes in result
	 * @see sqlite3_column_bytes()
	 */
	unsigned int column_bytes(
		unsigned int index );

	/**
	 * Get a value from the row
	 * @param index column index
	 * @param value reference to object to set with the value
	 * @see sqlite3_column_*()
	 */
	template< class T >
	void column(
		unsigned int index,
		T &value )
	{
		assert( index < _query.column_count() );
		const char *text = reinterpret_cast< const char * >(
			sqlite3_column_text( _query._handle, index ) );
		if( text )
			value = boost::lexical_cast< T >( text );
		else
			value = boost::get( boost::value_initialized< T >() );
	}

	/**
	 * Stream operator is used to obtain values from a result row, fetching from
	 * each column in turn. In addition, the null and set_index() auto-column-
	 * getting manipulators can be used.
	 * @param value is a variable to store the retrieved data in
	 */
	template< class T >
	row &operator >>(
		T &value )
	{
		column( _column_index, value );
		_column_index++;
		return *this;
	}

	/**
	 * Stream operator for use with set_index().
	 */
	row &operator >>(
		_set_index_t t );

	/**
	 * Check of this row is valid
	 * @return true if it is
	 */
	operator bool() const
	{
		return _valid;
	}

//______________________________________________________________________________
//                                                                implementation
protected:

	friend class query;

	/** the parent query */
	query &_query;

private:

	/** index used when auto-column-getting */
	unsigned int _column_index;

	/** is this row valid? */
	bool _valid;

};


// template specialisations
template< >
row &row::operator >> < _null_t >(
	_null_t & );


} // namespace row


#endif /* SQLITE3CC_ROW_H_ */
